@@ -1,4 +1,5 @@
 import json
+import logging
 import os
 import sys
 import time
@@ -18,8 +19,6 @@ from luigi import LuigiStatusCode
 from servicecatalog_puppet import config, constants
 from servicecatalog_puppet.workflow import tasks
 
-import logging
-
 logger = logging.getLogger()
 logger.setLevel(logging.INFO)
 
@@ -32,7 +31,6 @@ def run_tasks(
     is_dry_run=False,
     is_list_launches=None,
     execution_mode="hub",
-    cache_invalidator="now",
     on_complete_url=None,
     running_exploded=False,
 ):
@@ -105,6 +103,8 @@ def run_tasks(
         LuigiStatusCode.MISSING_EXT: 5,
     }
 
+    cache_invalidator = os.environ.get("SCT_CACHE_INVALIDATOR")
+
     if is_list_launches:
         if is_list_launches == "table":
             table = [
